/*
 * Tiny Vector Matrix Library
 * Dense Vector Matrix Libary of Tiny size using Expression Templates
 *
 * Copyright (C) 2001, 2002 Olaf Petzold <opetzold@wit.regiocom.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *	
 * "$Id:"
 */

#ifndef TVMET_VECTOR_OPERATORS_H
#define TVMET_VECTOR_OPERATORS_H

/* forwards */
#ifdef HAVE_COMPLEX
namespace std {
  template<class T> class complex;
}
#endif

namespace tvmet {

/**
 * \fn operator<<(std::ostream& os, const Vector<T, Sz>& rhs)
 * \brief Overload operator for i/o 
 * \ingroup _operator
 */
template<class T, std::size_t Sz>
inline
std::ostream& operator<<(std::ostream& os, const Vector<T, Sz>& rhs) {
  return rhs.printOn(os);
}

/***********************************************************************
 * global binary math, bitops and logical 
 ***********************************************************************/

/*
 * Vector<T1, Sz> and Vector<T2, Sz>
 */
#define DECLARE_BINARY_OPERATOR(FUNCTIONAL, OP)                   	\
template<class T1, class T2, std::size_t Sz> 	                  	\
inline                 			      			      	\
XprVector<                   						\
  XprBinOp<                   						\
    FUNCTIONAL<T1, T2>,                   				\
    VectorConstReference<T1, Sz>,              				\
    VectorConstReference<T2, Sz>               				\
  >,									\
  Sz									\
>                   			              			\
operator OP (const Vector<T1, Sz>& lhs, const Vector<T2, Sz>& rhs) { 	\
  typedef XprBinOp <                    				\
    FUNCTIONAL<T1, T2>,  	                 			\
    VectorConstReference<T1, Sz>,              				\
    VectorConstReference<T2, Sz>                     	      		\
  >							ExprType;	\
  return XprVector<ExprType, Sz>(ExprType(lhs.cref(), rhs.cref()));	\
}

DECLARE_BINARY_OPERATOR(fcnl_Add, +)
DECLARE_BINARY_OPERATOR(fcnl_Sub, -)
DECLARE_BINARY_OPERATOR(fcnl_Mul, *)
DECLARE_BINARY_OPERATOR(fcnl_Div, /)
DECLARE_BINARY_OPERATOR(fcnl_Mod, %)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseXor, ^)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseAnd, &)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseOr, |)
DECLARE_BINARY_OPERATOR(fcnl_Shl, <<)
DECLARE_BINARY_OPERATOR(fcnl_Shr, >>)
DECLARE_BINARY_OPERATOR(fcnl_Greater, >)
DECLARE_BINARY_OPERATOR(fcnl_Less, <)
DECLARE_BINARY_OPERATOR(fcnl_GreaterOrEqual, >=)
DECLARE_BINARY_OPERATOR(fcnl_LessOrEqual, <=)
DECLARE_BINARY_OPERATOR(fcnl_Equal, ==)
DECLARE_BINARY_OPERATOR(fcnl_NotEqual, !=)
DECLARE_BINARY_OPERATOR(fcnl_LogicalAnd, &&)
DECLARE_BINARY_OPERATOR(fcnl_LogicalOr, ||)
#undef DECLARE_BINARY_OPERATOR

/*
 * Vector<T, Sz> and POD's 
 */
#define DECLARE_BINARY_OPERATOR(FUNCTIONAL, OP, TP)               	\
template<class T, std::size_t Sz>              				\
inline                 			      			      	\
XprVector<             							\
  XprBinOp<          							\
    FUNCTIONAL< T, TP >,          					\
    VectorConstReference<T, Sz>,       					\
    XprLiteral< TP >          						\
  >,          								\
  Sz									\
>          								\
operator OP (const Vector<T, Sz>& lhs, TP rhs) {			\
  typedef XprBinOp<          						\
    FUNCTIONAL<T, TP >,          					\
    VectorConstReference<T, Sz>,       					\
    XprLiteral< TP >          						\
  >							ExprType;	\
  return XprVector<ExprType, Sz>(ExprType(lhs.cref(), XprLiteral< TP >(rhs))); \
}          								\
          								\
template<class T, std::size_t Sz>     					\
inline                 			      			      	\
XprVector<  								\
  XprBinOp<          							\
    FUNCTIONAL< TP, T>,          					\
    XprLiteral< TP >,          						\
    VectorConstReference<T, Sz>       					\
  >,          								\
  Sz									\
>          								\
operator OP (TP lhs, const Vector<T, Sz>& rhs) {         		\
  typedef XprBinOp<         						\
    FUNCTIONAL< TP, T>,         					\
    XprLiteral< TP >,         						\
    VectorConstReference<T, Sz>       					\
  >							ExprType;	\
  return XprVector<ExprType, Sz>(ExprType(XprLiteral< TP >(lhs), rhs.cref())); \
}

DECLARE_BINARY_OPERATOR(fcnl_Add, +, int)
DECLARE_BINARY_OPERATOR(fcnl_Sub, -, int)
DECLARE_BINARY_OPERATOR(fcnl_Mul, *, int)
DECLARE_BINARY_OPERATOR(fcnl_Div, /, int)
DECLARE_BINARY_OPERATOR(fcnl_Mod, %, int)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseXor, ^, int)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseAnd, &, int)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseOr, |, int)
DECLARE_BINARY_OPERATOR(fcnl_Shl, <<, int)
DECLARE_BINARY_OPERATOR(fcnl_Shr, >>, int)
DECLARE_BINARY_OPERATOR(fcnl_Greater, >, int)
DECLARE_BINARY_OPERATOR(fcnl_Less, <, int)
DECLARE_BINARY_OPERATOR(fcnl_GreaterOrEqual, >=, int)
DECLARE_BINARY_OPERATOR(fcnl_LessOrEqual, <=, int)
DECLARE_BINARY_OPERATOR(fcnl_Equal, ==, int)
DECLARE_BINARY_OPERATOR(fcnl_NotEqual, !=, int)
DECLARE_BINARY_OPERATOR(fcnl_LogicalAnd, &&, int)
DECLARE_BINARY_OPERATOR(fcnl_LogicalOr, ||, int)

DECLARE_BINARY_OPERATOR(fcnl_Add, +, float)
DECLARE_BINARY_OPERATOR(fcnl_Sub, -, float)
DECLARE_BINARY_OPERATOR(fcnl_Mul, *, float)
DECLARE_BINARY_OPERATOR(fcnl_Div, /, float)
DECLARE_BINARY_OPERATOR(fcnl_Greater, >, float)
DECLARE_BINARY_OPERATOR(fcnl_Less, <, float)
DECLARE_BINARY_OPERATOR(fcnl_GreaterOrEqual, >=, float)
DECLARE_BINARY_OPERATOR(fcnl_LessOrEqual, <=, float)
DECLARE_BINARY_OPERATOR(fcnl_Equal, ==, float)
DECLARE_BINARY_OPERATOR(fcnl_NotEqual, !=, float)
DECLARE_BINARY_OPERATOR(fcnl_LogicalAnd, &&, float)
DECLARE_BINARY_OPERATOR(fcnl_LogicalOr, ||, float)

DECLARE_BINARY_OPERATOR(fcnl_Add, +, double)
DECLARE_BINARY_OPERATOR(fcnl_Sub, -, double)
DECLARE_BINARY_OPERATOR(fcnl_Mul, *, double)
DECLARE_BINARY_OPERATOR(fcnl_Div, /, double)
DECLARE_BINARY_OPERATOR(fcnl_Greater, >, double)
DECLARE_BINARY_OPERATOR(fcnl_Less, <, double)
DECLARE_BINARY_OPERATOR(fcnl_GreaterOrEqual, >=, double)
DECLARE_BINARY_OPERATOR(fcnl_LessOrEqual, <=, double)
DECLARE_BINARY_OPERATOR(fcnl_Equal, ==, double)
DECLARE_BINARY_OPERATOR(fcnl_NotEqual, !=, double)
DECLARE_BINARY_OPERATOR(fcnl_LogicalAnd, &&, double)
DECLARE_BINARY_OPERATOR(fcnl_LogicalOr, ||, double)

#ifdef HAVE_LONG_DOUBLE
DECLARE_BINARY_OPERATOR(fcnl_Add, +, long double)
DECLARE_BINARY_OPERATOR(fcnl_Sub, -, long double)
DECLARE_BINARY_OPERATOR(fcnl_Mul, *, long double)
DECLARE_BINARY_OPERATOR(fcnl_Div, /, long double)
DECLARE_BINARY_OPERATOR(fcnl_Greater, >, long double)
DECLARE_BINARY_OPERATOR(fcnl_Less, <, long double)
DECLARE_BINARY_OPERATOR(fcnl_GreaterOrEqual, >=, long double)
DECLARE_BINARY_OPERATOR(fcnl_LessOrEqual, <=, long double)
DECLARE_BINARY_OPERATOR(fcnl_Equal, ==, long double)
DECLARE_BINARY_OPERATOR(fcnl_NotEqual, !=, long double)
DECLARE_BINARY_OPERATOR(fcnl_LogicalAnd, &&, long double)
DECLARE_BINARY_OPERATOR(fcnl_LogicalOr, ||, long double)
#endif // HAVE_LONG_DOUBLE

#undef DECLARE_BINARY_OPERATOR

#ifdef HAVE_COMPLEX

/*
 * Vector<T, Sz> and std::complex<>
 */
#define DECLARE_BINARY_OPERATOR(FUNCTIONAL, OP)		               	\
template<class T, std::size_t Sz>              				\
inline                 			      			      	\
XprVector<             							\
  XprBinOp<          							\
    FUNCTIONAL< T, std::complex<T> >,          				\
    VectorConstReference<T, Sz>,       					\
    XprLiteral< std::complex<T> >          				\
  >,          								\
  Sz									\
>          								\
operator OP (const Vector<T, Sz>& lhs, const std::complex<T>& rhs) {	\
  typedef XprBinOp<          						\
    FUNCTIONAL<T, std::complex<T> >,          				\
    VectorConstReference<T, Sz>,       					\
    XprLiteral< std::complex<T> >          				\
  >							ExprType;	\
  return XprVector<ExprType, Sz>(ExprType(lhs.cref(), XprLiteral< std::complex<T> >(rhs))); \
}          								\
          								\
template<class T, std::size_t Sz>     					\
inline                 			      			      	\
XprVector<  								\
  XprBinOp<          							\
    FUNCTIONAL< std::complex<T>, T>,          				\
    XprLiteral< std::complex<T> >,          				\
    VectorConstReference<T, Sz>       					\
  >,          								\
  Sz									\
>          								\
operator OP (const std::complex<T>& lhs, const Vector<T, Sz>& rhs) { 	\
  typedef XprBinOp<         						\
    FUNCTIONAL< std::complex<T>, T>,         				\
    XprLiteral< std::complex<T> >,         				\
    VectorConstReference<T, Sz>       					\
  >							ExprType;	\
  return XprVector<ExprType, Sz>(ExprType(XprLiteral< std::complex<T> >(lhs), rhs.cref())); \
}

DECLARE_BINARY_OPERATOR(fcnl_Add, +)
DECLARE_BINARY_OPERATOR(fcnl_Sub, -)
DECLARE_BINARY_OPERATOR(fcnl_Mul, *)
DECLARE_BINARY_OPERATOR(fcnl_Div, /)
DECLARE_BINARY_OPERATOR(fcnl_Mod, %)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseXor, ^)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseAnd, &)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseOr, |)
DECLARE_BINARY_OPERATOR(fcnl_Shl, <<)
DECLARE_BINARY_OPERATOR(fcnl_Shr, >>)
DECLARE_BINARY_OPERATOR(fcnl_Greater, >)
DECLARE_BINARY_OPERATOR(fcnl_Less, <)
DECLARE_BINARY_OPERATOR(fcnl_GreaterOrEqual, >=)
DECLARE_BINARY_OPERATOR(fcnl_LessOrEqual, <=)
DECLARE_BINARY_OPERATOR(fcnl_Equal, ==)
DECLARE_BINARY_OPERATOR(fcnl_NotEqual, !=)
DECLARE_BINARY_OPERATOR(fcnl_LogicalAnd, &&)
DECLARE_BINARY_OPERATOR(fcnl_LogicalOr, ||)
#undef DECLARE_BINARY_OPERATOR

#endif // HAVE_COMPLEX

/*
 * XprVector<E, Sz> and Vector<T, Sz> 
 */
#define DECLARE_BINARY_OPERATOR(FUNCTIONAL, OP)                   	\
template<class E, class T, std::size_t Sz>             			\
inline                 			      			      	\
XprVector<                    	      					\
  XprBinOp<                     	      				\
    FUNCTIONAL<typename E::value_type, T>,                    	      	\
    XprVector<E, Sz>,                  	      				\
    VectorConstReference<T, Sz>                	      			\
  >,                     	      					\
  Sz									\
>                      	      						\
operator OP (XprVector<E, Sz> lhs, const Vector<T, Sz>& rhs) {		\
  typedef XprBinOp<                     	      			\
    FUNCTIONAL<typename E::value_type, T>,                    	      	\
    XprVector<E, Sz>,                  	      				\
    VectorConstReference<T, Sz>                	      			\
  > 							 ExprType; 	\
  return XprVector<ExprType, Sz>(ExprType(lhs, rhs.cref()));		\
}                    	      						\
                    	      						\
template<class E, class T, std::size_t Sz>             			\
inline                 			      			      	\
XprVector<                    	      					\
  XprBinOp<                     	      				\
    FUNCTIONAL<T, typename E::value_type>,                    	      	\
    VectorConstReference<T, Sz>,               	      			\
    XprVector<E, Sz>                   	      				\
  >,                     	      					\
  Sz									\
>                      	      						\
operator OP (const Vector<T, Sz>& lhs, XprVector<E, Sz> rhs) {         	\
  typedef XprBinOp<                     	      			\
    FUNCTIONAL<T, typename E::value_type>,                    	      	\
    VectorConstReference<T, Sz>,      	      				\
    XprVector<E, Sz>                  	      				\
  > 						 	ExprType; 	\
  return XprVector<ExprType, Sz>(ExprType(lhs.cref(), rhs));         	\
}

DECLARE_BINARY_OPERATOR(fcnl_Add, +)
DECLARE_BINARY_OPERATOR(fcnl_Sub, -)
DECLARE_BINARY_OPERATOR(fcnl_Mul, *)
DECLARE_BINARY_OPERATOR(fcnl_Div, /)
DECLARE_BINARY_OPERATOR(fcnl_Mod, %)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseXor, ^)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseAnd, &)
DECLARE_BINARY_OPERATOR(fcnl_BitwiseOr, |)
DECLARE_BINARY_OPERATOR(fcnl_Shl, <<)
DECLARE_BINARY_OPERATOR(fcnl_Shr, >>)
DECLARE_BINARY_OPERATOR(fcnl_Greater, >)
DECLARE_BINARY_OPERATOR(fcnl_Less, <)
DECLARE_BINARY_OPERATOR(fcnl_GreaterOrEqual, >=)
DECLARE_BINARY_OPERATOR(fcnl_LessOrEqual, <=)
DECLARE_BINARY_OPERATOR(fcnl_Equal, ==)
DECLARE_BINARY_OPERATOR(fcnl_NotEqual, !=)
DECLARE_BINARY_OPERATOR(fcnl_LogicalAnd, &&)
DECLARE_BINARY_OPERATOR(fcnl_LogicalOr, ||)
#undef DECLARE_BINARY_OPERATOR

/***********************************************************************
 * global unary operators
 ***********************************************************************/

/*
 * Operators on Vector<T, Sz>
 */
#define DECLARE_UNARY_OPERATOR(FUNCTIONAL, OP)                   	\
template <class T, std::size_t Sz>                                      \
inline                                                        		\
XprVector<                                                    		\
  XprUnOp<                                                        	\
    FUNCTIONAL<T>,                                                      \
    VectorConstReference<T, Sz>                                       	\
  >,                                                        		\
  Sz									\
>                                                        		\
operator OP (const Vector<T, Sz>& rhs) {                                \
  typedef XprUnOp<                                                      \
    FUNCTIONAL<T>,                                                      \
    VectorConstReference<T, Sz>                                        	\
  >  							 ExprType;	\
  return XprVector<ExprType, Sz>(ExprType(rhs.cref()));              	\
}

DECLARE_UNARY_OPERATOR(fcnl_LogicalNot, !)
DECLARE_UNARY_OPERATOR(fcnl_BitwiseNot, ~)
DECLARE_UNARY_OPERATOR(fcnl_Negate, -)
#undef DECLARE_UNARY_OPERATOR

} // namespace tvmet

#endif // TVMET_VECTOR_OPERATORS_H

// Local Variables:
// mode:C++
// End:
